
from mlmnemonist.experiment_runner import ExperimentRunner
from mlmnemonist.experiment_factory import ExperimentRunnerFactory
from mlmnemonist.runner_cache import MAX_CACHE_SIZE

__version__ = '0.0.9'


from typing import List

import torch
import torch.nn as nn
import torch.nn.functional as F
from testing.modeling.registry import Registry
from yacs.config import CfgNode as ConfigurationNode

FUSION_REGISTRY = Registry()


class DrugConcentrationFusion(nn.Module):
    def __init__(self, conc_in_features, drug_in_features):
        super(DrugConcentrationFusion, self).__init__()
        self.conc_in_features = conc_in_features
        self.drug_in_features = drug_in_features
        self.out_features = None


# Fusions using MLP and concatenation
class MLPBasedFusion(DrugConcentrationFusion):
    def __init__(self, conc_in_features: int, drug_in_features: int,
                 hidden_units: List[int],
                 dropout: float):
        super(MLPBasedFusion, self).__init__(conc_in_features, drug_in_features)
        prv = conc_in_features + drug_in_features
        for i, h in enumerate(hidden_units):
            self.add_module(f'fc_{i + 1}', nn.Linear(prv, h))
            prv = h
            if i % 2 == 0:
                self.add_module(f'batchnorm_{i + 1}', nn.BatchNorm1d(h))
        self.dropout_rate = dropout
        self.layer_count = len(hidden_units)
        self.out_features = prv


def _custom_forward(module: MLPBasedFusion, x):
    for i in range(module.layer_count):
        L = module._modules[f'fc_{i + 1}']
        x = L(x)
        x = F.relu(x)
        if i % 2 == 0:
            L = module._modules[f'batchnorm_{i + 1}']
            x = L(x)
        if i < module.layer_count - 1:
            x = F.dropout(x, p=module.dropout_rate, training=module.training)
    return x


class ConcatMLPConcentrationFusion(MLPBasedFusion):
    def __init__(self, conc_in_features: int, drug_in_features: int, hidden_units: List[int], dropout: float):
        super(ConcatMLPConcentrationFusion, self).__init__(conc_in_features, drug_in_features, hidden_units, dropout)

    def forward(self, drugs, concentrations):
        # drugs : [B x drug_embeddings]
        # concentrations : [B x concentration embeddings]
        x = torch.cat([drugs, concentrations], axis=1)
        return _custom_forward(self, x)


class ConcatMLPConcentrationFusionNoCompress(MLPBasedFusion):
    def __init__(self, conc_in_features: int, drug_in_features: int, hidden_units: List[int], dropout: float):
        super(ConcatMLPConcentrationFusionNoCompress, self).__init__(conc_in_features, drug_in_features, hidden_units,
                                                                     dropout)

    def forward(self, compound_nodes: torch.Tensor, batch_index: torch.Tensor, concentrations: torch.Tensor) -> [
        torch.Tensor, torch.Tensor]:
        # compound_nodes -> All the nodes of the compound collated
        # batch_index -> An index array the size of compound_nodes.shape[0] which defines
        # the corresponding batch of each node
        # Concentrations -> [B x conc_in_features]
        all_conc_nodes = concentrations[batch_index, :]
        joined = torch.cat([compound_nodes, all_conc_nodes], axis=1)
        ret = _custom_forward(self, joined)
        return ret, batch_index


@FUSION_REGISTRY.register('mlp-concat-fusion')
def build_mlp_concant_fusion(cfg_hyper_parameters: ConfigurationNode, drug_in_features: int):
    return ConcatMLPConcentrationFusion(
        dropout=cfg_hyper_parameters.DROPOUT,
        conc_in_features=cfg_hyper_parameters.CONC_IN_FEATURES,
        drug_in_features=drug_in_features,
        hidden_units=cfg_hyper_parameters.HIDDEN_UNITS
    )


@FUSION_REGISTRY.register('mlp-concat-fusion-no-compression')
def build_mlp_concat_fusion_with_no_compress(cfg_hyper_parameters: ConfigurationNode, drug_in_features: int):
    return ConcatMLPConcentrationFusionNoCompress(
        dropout=cfg_hyper_parameters.DROPOUT,
        conc_in_features=cfg_hyper_parameters.CONC_IN_FEATURES,
        drug_in_features=drug_in_features,
        hidden_units=cfg_hyper_parameters.HIDDEN_UNITS
    )




"""
A config file drawn from another project that can be used as a testing palette here
"""
from yacs.config import CfgNode


class ConfigurationNode(CfgNode):
    """
    Same as CfgNode but with a hashing ability added for caching!
    """
    def __hash__(self):
        return self.__str__().__hash__()


# YACS overwrite these settings using YAML
__C = ConfigurationNode()

__C.MODEL_HAS_CHECKPOINTS = True

# The description for the experiment that is being done
__C.DESCRIPTION = 'no description!'

__C.DATASET = ConfigurationNode()

# Setup dataset settings
__C.DATASET.PATH_DATA_RAW = ConfigurationNode()
__C.DATASET.PATH_DATA_RAW.COMBO = '/path/not/specified/'
__C.DATASET.PATH_DATA_RAW.INFO = '/path/not/specified/'

#
__C.DATASET.COMBO_TABLE_NAME = 'ALMANAC2017.csv'
__C.DATASET.HAS_STUDY_SOURCES = False
__C.DATASET.STUDY_SOURCES = ['ALMANAC']
__C.DATASET.USE_EXTERNAL_INDICES_DIR = False
__C.DATASET.EXTERNAL_INDICES_DIR = '/path/not/specified/'
__C.DATASET.USE_EXTERNAL_INDICES = False
__C.DATASET.EXTERNAL_INDICES = []
__C.DATASET.RESPONSE_UPPER_BOUND = 500
__C.DATASET.RESPONSE_LOWER_BOUND = -500
__C.DATASET.USE_FOCUS_DRUG_LIST = False
__C.DATASET.FOCUS_DRUG_LIST = []
__C.DATASET.USE_FOCUS_CELL_LIST = False
__C.DATASET.FOCUS_CELL_LIST = []
__C.DATASET.USE_SCREENING_SAMPLE_RATIO = False
__C.DATASET.SCREENING_SAMPLE_RATIO = 1.0
__C.DATASET.USE_SCREENING_SAMPLE_COUNT = False
__C.DATASET.SCREENING_SAMPLE_COUNT = 4
__C.DATASET.CONSIDER_ONLY_MONOTHERAPIES = False

# Setup dataset drug setting
__C.DATASET.DRUG = ConfigurationNode()
__C.DATASET.DRUG.USE_SMILES = False
__C.DATASET.DRUG.GUIDE = 'not specified!'
__C.DATASET.DRUG.ATOM_FEATURES = 78
__C.DATASET.DRUG.USE_DRUG_EMBEDDINGS = True
__C.DATASET.DRUG.DETAIL_LEVEL = 1
__C.DATASET.DRUG.INCLUDE_MACCS = False
# Setup dataset cell setting
__C.DATASET.CELL = ConfigurationNode()
__C.DATASET.CELL.USE_EXPRESSION = True
__C.DATASET.CELL.EXPRESSION_SOURCE = 'ccle'
__C.DATASET.CELL.USE_EXPRESSION_ENCODER = False
__C.DATASET.CELL.EXPRESSION_ENCODER_TYPE = 'generic'
__C.DATASET.CELL.USE_GENE = False
__C.DATASET.CELL.REPLACE_GENE_NAMES_WITH_IDS = False
__C.DATASET.CELL.INCLUDE_CODE = True
__C.DATASET.CELL.INCLUDE_CNV = True
__C.DATASET.CELL.INCLUDE_CRISPR = True
__C.DATASET.CELL.INCLUDE_PROTEOMICS = True
__C.DATASET.CELL.DETAIL_LEVEL = 'D1'
# Setup dataset concentration setting
__C.DATASET.CONCENTRATION = ConfigurationNode()
__C.DATASET.CONCENTRATION.USE_ENCODER = True
__C.DATASET.CONCENTRATION.ENCODER_TYPE = 'one-hot'

# Setup training and testing splits
__C.TRAIN_TEST_SPLIT = ConfigurationNode()
__C.TRAIN_TEST_SPLIT.SPLIT_DIR = '/path/not/specified/'
__C.TRAIN_TEST_SPLIT.FOLD_INDICES_FILENAME = 'name_not_specified.pkl'

###
# Solving method
# - DeepDDR
# - Random Forest
###
__C.SOLVING_METHOD = 'Random Forest'

##
# Random forest configs
__C.RANDOM_FOREST = ConfigurationNode()

## Random forest dataloader configs
__C.RANDOM_FOREST.DATA_LOADER = ConfigurationNode()
__C.RANDOM_FOREST.DATA_LOADER.USE_DRUG_EMBEDDING = True
__C.RANDOM_FOREST.DATA_LOADER.USE_CONC_EMBEDDING = True
__C.RANDOM_FOREST.DATA_LOADER.USE_CELL_EMBEDDING = True
__C.RANDOM_FOREST.DATA_LOADER.USE_EXPRESSION = True

## Random forest solver settings
__C.RANDOM_FOREST.MODEL = ConfigurationNode()
__C.RANDOM_FOREST.MODEL.N_ESTIMATOR = 100

__C.RANDOM_FOREST.SOLVER = ConfigurationNode()
__C.RANDOM_FOREST.SOLVER.BATCH_SIZE = 5000
__C.RANDOM_FOREST.SOLVER.NUM_OF_WORKERS = -1  # Set to maximum

##
# Deep DDRM method
__C.DEEP_DDR = ConfigurationNode()
__C.DEEP_DDR.MODEL = ConfigurationNode()
__C.DEEP_DDR.MODEL.TYPE = 'type not specified!'
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS = ConfigurationNode()
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.DROPOUT = 0.0

# Setup drug embedder
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.DRUG_EMBEDDER = ConfigurationNode()
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.DRUG_EMBEDDER.TYPE = 'type not specified'
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.DRUG_EMBEDDER.HYPER_PARAMETERS = ConfigurationNode()
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.DRUG_EMBEDDER.HYPER_PARAMETERS.GNN_TYPE = 'GNN not specified!'
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.DRUG_EMBEDDER.HYPER_PARAMETERS.IN_FEATURES = -1
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.DRUG_EMBEDDER.HYPER_PARAMETERS.DROPOUT = 0.0
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.DRUG_EMBEDDER.HYPER_PARAMETERS.D_LAYERS = (-1, -1)
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.DRUG_EMBEDDER.HYPER_PARAMETERS.OUT_FEATURES = -1
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.DRUG_EMBEDDER.HYPER_PARAMETERS.ACTIVATION = 'activation not specified!'
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.DRUG_EMBEDDER.HYPER_PARAMETERS.N_HEADS = -1
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.DRUG_EMBEDDER.HYPER_PARAMETERS.POOLING_OUT_FEATURES = -1

# Setup cell embedder
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.CELL_EMBEDDER = ConfigurationNode()
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.CELL_EMBEDDER.TYPE = 'type not specified'
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.CELL_EMBEDDER.HYPER_PARAMETERS = ConfigurationNode()
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.CELL_EMBEDDER.HYPER_PARAMETERS.DROPOUT = 0.0
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.CELL_EMBEDDER.HYPER_PARAMETERS.IN_FEATURES = -1
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.CELL_EMBEDDER.HYPER_PARAMETERS.OUT_FEATURES = -1
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.CELL_EMBEDDER.HYPER_PARAMETERS.H1 = -1
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.CELL_EMBEDDER.HYPER_PARAMETERS.H2 = -1

# Setup Fusion model
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.FUSION = ConfigurationNode()
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.FUSION.TYPE = 'type not specified'
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.FUSION.HYPER_PARAMETERS = ConfigurationNode()
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.FUSION.HYPER_PARAMETERS.DROPOUT = 0.0
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.FUSION.HYPER_PARAMETERS.CONC_IN_FEATURES = -1
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.FUSION.HYPER_PARAMETERS.HIDDEN_UNITS = (-1, -1)

# Setup integration model
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.INTEGRATION = ConfigurationNode()
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.INTEGRATION.TYPE = 'type not specified'
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.INTEGRATION.HYPER_PARAMETERS = ConfigurationNode()
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.INTEGRATION.HYPER_PARAMETERS.MAX_DRUG_COUNT = -1
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.INTEGRATION.HYPER_PARAMETERS.N_HEADS = -1
__C.DEEP_DDR.MODEL.HYPER_PARAMETERS.INTEGRATION.HYPER_PARAMETERS.H = -1

# The solver parameters
__C.DEEP_DDR.DATA_LOADER = ConfigurationNode()
__C.DEEP_DDR.DATA_LOADER.BATCH_SIZE = -1

__C.DEEP_DDR.SOLVER = ConfigurationNode()
__C.DEEP_DDR.SOLVER.USE_PRETRAINED_WEIGHTS = False
__C.DEEP_DDR.SOLVER.WEIGHT_INITIALIZATION_DIR = '/path/not/specified/'
__C.DEEP_DDR.SOLVER.N_EPOCHS = -1
__C.DEEP_DDR.SOLVER.DEVICE = 'device not specified!'
__C.DEEP_DDR.SOLVER.LOSS = 'loss not specified!'
__C.DEEP_DDR.SOLVER.OPTIMIZER = ConfigurationNode()
__C.DEEP_DDR.SOLVER.OPTIMIZER.TYPE = 'device not specified!'
__C.DEEP_DDR.SOLVER.OPTIMIZER.LR = 0.0

__C.OUTPUT = ConfigurationNode()
__C.OUTPUT.FILES_DIR = ['/path/not/specified1/', '/path/not/specified2/']
__C.OUTPUT.DESCRIPTIONS = ['description not specified1', 'description not specified2']


def get_cfg_defaults():
    """
    Get a yacs CfgNode object with default values
    """
    # Return a clone so that the defaults will not be altered
    # It will be subsequently overwritten with local YAML.
    return __C.clone()



# -*- coding: UTF-8 -*-

import io
import json
from typing import Optional

from openpyxl.workbook import Workbook

import drive
from drive import mimetypes


class File:
    """
    A file on Google Drive. This might be a directory as well.
    """

    def __init__(self, attrs, client: Optional["drive.Client"] = None):
        """

        :param attrs:
        :param client:
        """
        # If one calls File(File(...)) make the outer file copy the attributes of the inner one
        if isinstance(attrs, File):
            for attr in ("id", "_name", "kind", "mimetype", "size", "parents", "_client"):
                setattr(self, attr, getattr(attrs, attr))
            return

        self.id = attrs["id"]
        self._name = attrs.get("name")
        self.kind = attrs.get("kind")
        self.mimetype = attrs.get("mimeType")
        self.parents_ids = attrs.get("parents")

        self.size = attrs.get("size")
        if self.size:
            self.size = int(self.size)

        self._client = client

    @property
    def name(self):
        if self._name is None and self._client:
            me = self._client.get_file_metadata(self.id, False)
            if me:
                self._name = me["name"]

        return self._name

    @property
    def is_directory(self):
        if self.mimetype is None:
            return None
        return self.mimetype == mimetypes.GOOGLE_DRIVE_FOLDER

    @property
    def human_type(self):
        """
        Human-friendly file type.
        :return:
        """
        aliases = {
            mimetypes.GOOGLE_DRIVE_FOLDER: "folder",
            mimetypes.GOOGLE_SHEETS: "Google Spreadsheet",
            mimetypes.JSON: "JSON",
        }

        if self.mimetype in aliases:
            return aliases[self.mimetype]

        if self.name and self.name.lower().endswith(".jsons"):
            return "JSONS"

        return self.mimetype or "?"

    def exists(self):
        """
        Test if the file exists.
        """
        if not self._client:
            return None

        return bool(self._client.get_file(self.id, raise_if_not_found=False))

    def unlink(self):
        """
        Remove the file. If it's a directory all its children are removed as well.
        :return: a boolean indicating success
        """
        if not self._client:
            return False

        # "If successful, this method returns an empty response body."
        # https://developers.google.com/drive/v3/reference/files/delete
        return self._client.remove_file(self.id) == ""

    def rename(self, new_name):
        """
        Rename the file.
        :param new_name:
        :return:
        """
        if not self._client:
            return False
        m = self._client.update_file(self.id, name=new_name)
        self._update(m)

    def move_in(self, new_parent, new_name=None):
        """
        Move the file under a new parent.
        :param new_parent:
        :param new_name:
        :return:
        """
        if new_parent.is_directory is False:
            raise NotADirectoryError(new_parent.name)

        parents_ids = [new_parent.id]
        kw = {
            "add_parents_ids": parents_ids,
            "remove_parents_ids": [p.id for p in self.parents()],
        }
        if new_name:
            kw["name"] = new_name

        m = self._client.update_file(self.id, **kw)
        self._update(m)
        self.parents_ids = parents_ids

    def list(self):
        """
        List a directory's content. This returns an empty list for simple files.
        :return:
        """
        if not self._client or not self.is_directory:
            return []

        return self._client.list_files(parents_in=self.id)

    def create_folder(self, name):
        """
        Create a folder under a directory. This has no effect if the file is not a directory.
        :param name:
        :return:
        """
        if not self._client or not self.is_directory:
            return False

        return self._client.create_folder(name, self.id)

    def get_or_create_folder(self, name):
        """

        :param name:
        :return:
        """
        if not self._client or not self.is_directory:
            return False

        return self._client.get_or_create_folder(name, self.id)

    def grant_permissions(self, role: str, type_: str):
        if not self._client:
            return None

        return self._client.grant_file_permissions(self.id, role, type_)

    def get_child(self, name):
        """
        Get a child file. Return None if the current file is not a directory.

        :param name:
        :return:
        """
        if not self._client or not self.is_directory:
            return None

        return self._client.get_file(name, self.id)

    def parents(self):
        """
        Return all parents of a file.
        Note that a file can have multiple parents on Google Drive.
        :return:
        """
        if self.parents_ids is None:
            if not self._client:
                return []

            m = self._client.get_file_metadata(self.id, fields="parents")
            if not m:
                self.parents_ids = []
                return []

            self.parents_ids = m["parents"]

        parents = []
        for pid in self.parents_ids:
            parents.append(File({
                "id": pid,
                "mimetype": mimetypes.GOOGLE_DRIVE_FOLDER,
            }, client=self._client))

        return parents

    def parent(self):
        """
        Return the first parent of a file.
        :return:
        """
        ps = self.parents()
        if not ps:
            return None
        return ps[0]

    def download(self, writer, mime_type=None):
        """

        :param writer:
        :param mime_type:
        :return:
        """
        if not self._client:
            return False
        return self._client.download(self.id, writer, mime_type=mime_type)

    def download_file(self, path, mime_type=None):
        """

        :param path:
        :param mime_type:
        :return:
        """
        if not self._client:
            return None
        return self._client.download_file(self.id, path, mime_type=mime_type)

    def download_workbook(self, read_only=False) -> Optional[Workbook]:
        """

        :param read_only: set this to ``True`` if you don't plan to save or edit the workbook.
        :return:
        """
        if not self._client:
            return None
        return self._client.download_excel_workbook(self.id, read_only=read_only)

    def get_bytes(self, mime_type=None):
        """
        Return a ``io.BytesIO`` object holding the content of the file. This can be used as a binary reader.
        :param mime_type:
        :return:
        """
        fh = io.BytesIO()
        self.download(fh, mime_type=mime_type)
        fh.seek(0)
        return fh

    def json(self):
        """
        Convenient method to download a file as JSON.
        :return:
        """
        return json.load(self.get_bytes())

    def jsons(self):
        """
        Convenient method to download a file as a newline-delimited list of JSON objects. This returns a generator.
        :return:
        """
        line = True
        b = self.get_bytes()
        while line:
            line = b.readline()
            if line:  # skip empty lines
                yield json.loads(line.decode("utf-8"))

    def get_web_view_link(self):
        return self._client.get_web_view_link(self.id)

    def to_dict(self):
        """
        Return the file’s metadata as a dict. Mandatory keys: ``"id"``, ``"name"``. Optional ones: ``"parents"``.
        :return:
        """
        d = {
            # Add other fields as needed
            "id": self.id,
            "name": self._name,
        }
        if self.parents_ids:
            d["parents"] = self.parents_ids

        return d

    def __str__(self):
        return self.name

    def __repr__(self):
        klass = self.__class__
        dir_info = " (directory)" if self.is_directory else ""
        return "<%s.%s name=\"%s\"%s>" % (
            klass.__module__, klass.__name__, self.__str__(), dir_info)

    # Allow files to be sorted
    def __lt__(self, other):
        return str(self) < str(other)

    # Private API

    def _update(self, attrs):
        """
        Update the current file instance with an attributes dict.
        :param attrs:
        :return:
        """
        if not attrs:
            return
        if isinstance(attrs, File):
            attrs = attrs.to_dict()

        # Add other fields as needed
        name = attrs.get("name")
        if name:
            self._name = name

        parents = attrs.get("parents")
        if parents:
            self.parents_ids = parents



"""Retain apiclient as an alias for googleapiclient."""

from six import iteritems

import googleapiclient

from googleapiclient import channel
from googleapiclient import discovery
from googleapiclient import errors
from googleapiclient import http
from googleapiclient import mimeparse
from googleapiclient import model

try:
    from googleapiclient import sample_tools
except ImportError:
    # Silently ignore, because the vast majority of consumers won't use it and
    # it has deep dependence on oauth2client, an optional dependency.
    sample_tools = None
from googleapiclient import schema

_SUBMODULES = {
    "channel": channel,
    "discovery": discovery,
    "errors": errors,
    "http": http,
    "mimeparse": mimeparse,
    "model": model,
    "sample_tools": sample_tools,
    "schema": schema,
}

import sys

for module_name, module in iteritems(_SUBMODULES):
    sys.modules["apiclient.%s" % module_name] = module



# -*- coding: UTF-8 -*-

import os
import httplib2
from os import environ
from typing import Optional

from oauth2client.service_account import ServiceAccountCredentials

from drive.exceptions import DriveException

ENV_CLIENT_SECRET_PATH = "GOOGLE_APPLICATION_CREDENTIALS"

# If you modify these scopes, delete your previously saved credentials
DRIVE_SCOPE = 'https://www.googleapis.com/auth/drive'


class MissingCredentialsException(DriveException):
    def __init__(self):
        super().__init__("Missing credentials! Please set %s" % ENV_CLIENT_SECRET_PATH)


def get_credentials(path: Optional[str] = None):
    """
    Retrieve the user's Google Cloud credentials.
    :param path:
    :return:
    """

    if path is None:
        path = environ["GOOGLE_APPLICATION_CREDENTIALS"]

    path = os.path.expanduser(path)
    creds = ServiceAccountCredentials.from_json_keyfile_name(path, scopes=[DRIVE_SCOPE])
    if not creds or creds.invalid:
        raise MissingCredentialsException()
    return creds


def authorize(credentials):
    """
    :param credentials:
    :return:
    """
    return credentials.authorize(httplib2.Http())



from setuptools import setup, find_packages

classfiers = [
    'Development Status :: In production :: Pre-release',
    'Programming language :: Python :: 3.8',
    'Operating System :: OS independent'
]

setup(
    name='mlmnemonist',
    version='0.0.9',
    description='A light-weight framework to store the progress you made'
                'on your ML operations with the ability to smartly cache your models'
                'and retrieve it even when your session crashes.',
    long_description=open('README.txt').read() + '\n\n' + open('CHANGELOG.txt').read(),
    url='https://github.com/HamidrezaKmK/ML-Mnemonist',
    author='Hamidreza Kamkari',
    author_email='hamidrezakamkari@gmail.com',
    license='MIT',
    classfiers=classfiers,
    keywords='MLOps, Caching',
    packages=find_packages(),
    install_requires=['python-dotenv', 'yacs']
)


# -*- coding: utf-8 -*-

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/type/datetime.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1agoogle/type/datetime.proto\x12\x0bgoogle.type\x1a\x1egoogle/protobuf/duration.proto"\xe0\x01\n\x08\x44\x61teTime\x12\x0c\n\x04year\x18\x01 \x01(\x05\x12\r\n\x05month\x18\x02 \x01(\x05\x12\x0b\n\x03\x64\x61y\x18\x03 \x01(\x05\x12\r\n\x05hours\x18\x04 \x01(\x05\x12\x0f\n\x07minutes\x18\x05 \x01(\x05\x12\x0f\n\x07seconds\x18\x06 \x01(\x05\x12\r\n\x05nanos\x18\x07 \x01(\x05\x12/\n\nutc_offset\x18\x08 \x01(\x0b\x32\x19.google.protobuf.DurationH\x00\x12*\n\ttime_zone\x18\t \x01(\x0b\x32\x15.google.type.TimeZoneH\x00\x42\r\n\x0btime_offset"\'\n\x08TimeZone\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\tBi\n\x0f\x63om.google.typeB\rDateTimeProtoP\x01Z<google.golang.org/genproto/googleapis/type/datetime;datetime\xf8\x01\x01\xa2\x02\x03GTPb\x06proto3'
)


_DATETIME = DESCRIPTOR.message_types_by_name["DateTime"]
_TIMEZONE = DESCRIPTOR.message_types_by_name["TimeZone"]
DateTime = _reflection.GeneratedProtocolMessageType(
    "DateTime",
    (_message.Message,),
    {
        "DESCRIPTOR": _DATETIME,
        "__module__": "google.type.datetime_pb2"
        # @@protoc_insertion_point(class_scope:google.type.DateTime)
    },
)
_sym_db.RegisterMessage(DateTime)

TimeZone = _reflection.GeneratedProtocolMessageType(
    "TimeZone",
    (_message.Message,),
    {
        "DESCRIPTOR": _TIMEZONE,
        "__module__": "google.type.datetime_pb2"
        # @@protoc_insertion_point(class_scope:google.type.TimeZone)
    },
)
_sym_db.RegisterMessage(TimeZone)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n\017com.google.typeB\rDateTimeProtoP\001Z<google.golang.org/genproto/googleapis/type/datetime;datetime\370\001\001\242\002\003GTP"
    _DATETIME._serialized_start = 76
    _DATETIME._serialized_end = 300
    _TIMEZONE._serialized_start = 302
    _TIMEZONE._serialized_end = 341
# @@protoc_insertion_point(module_scope)



import copy
import json
import os
from typing import Optional, List, Dict

from mlmnemonist.runner_cache import RunnerCache
from mlmnemonist import ExperimentRunner
from mlmnemonist.validation_tools import expand_cfg
from yacs.config import CfgNode as ConfigurationNode

import functools

GRID_SEARCH_CACHING = 500


def _convert(code: str) -> List[int]:
    return [int(x) for x in code.split('-')]


def _rev_convert(code: List[int]) -> str:
    return '-'.join([str(x) for x in code])


def _get_all_codes(parent_directory: str) -> Dict[str, str]:
    all_codes = {}
    for f in os.listdir(parent_directory):
        if f.split('.')[-1] == 'yaml':
            all_codes[f.split('-MLM-')[0]] = os.path.join(parent_directory, f)

    return all_codes


def _get_adjacent_based_on_list(curr_list: List[int], guide: List[int], curr_index: int = 0) -> List[List[int]]:
    if curr_index == len(curr_list):
        return [copy.deepcopy(curr_list)]
    ret = []

    branch1 = copy.deepcopy(curr_list)
    branch1[curr_index] += 1
    if branch1[curr_index] == guide[curr_index]:
        branch1[curr_index] = 0
    if branch1[curr_index] != curr_list[curr_index]:
        ret += _get_adjacent_based_on_list(branch1, guide, curr_index + 1)

    branch2 = copy.deepcopy(curr_list)
    branch2[curr_index] -= 1
    if branch2[curr_index] == -1:
        branch2[curr_index] = guide[curr_index] - 1
    if branch2[curr_index] != curr_list[curr_index]:
        ret += _get_adjacent_based_on_list(branch2, guide, curr_index + 1)

    ret += _get_adjacent_based_on_list(curr_list, guide, curr_index + 1)

    return ret


@functools.lru_cache(maxsize=GRID_SEARCH_CACHING)
def _get_adjacent_codes(code: str, all_codes: str) -> List[str]:
    """
    :param code:
    A code made up of digits separated by '-'

    :param all_codes:
    All the possible codes in our grid world

    :return:
    A list of codes adjacent to the current code
    """
    all_codes = all_codes.split('.')
    curr_list = [int(x) for x in code.split('-')]
    guide = []
    for code in all_codes:
        list_code = _convert(code)
        for i, x in enumerate(list_code):
            if i < len(guide):
                guide[i] = max(guide[i], x + 1)
            else:
                guide.append(x + 1)

    all_adj = _get_adjacent_based_on_list(curr_list, guide)
    all_adj = [_rev_convert(x) for x in all_adj]
    all_adj = [x for x in all_adj if x in all_codes]
    return all_adj


def _run_grid_search(runner: ExperimentRunner,
                     cache: RunnerCache,
                     verbose: int,
                     cfg_base: ConfigurationNode,
                     all_cfg_dir: str,
                     *args, **kwargs):
    cache.LOAD()
    # Add configurations
    runner.add_config(cfg_base=cfg_base)

    # Calculate the value for all iterations
    cfg_dict = _get_all_codes(all_cfg_dir)
    all_codes = list(cfg_dict.keys())
    curr = cache.SET_IFN('current_code', all_codes[0])
    marks = cache.SET_IFN('marks', set(list([curr])))
    score_dict = cache.SET_IFN('score_dict', {})
    iteration_i = cache.SET_IFN('iteration_i', 0)

    last_score = None
    while True:
        found_more = False
        for adjacent_code in _get_adjacent_codes(curr, '.'.join(all_codes)):
            if adjacent_code not in marks:
                runner.verbose = max(0, verbose - 1)
                runner.CACHE.RESET(prompt=False)
                runner.cfg_path = cfg_dict[adjacent_code]

                if verbose > 0:
                    print(f"Iteration no. [{iteration_i + 1}/{len(all_codes)}] "
                          f"-- Running {os.path.split(runner.cfg_path)[-1]}", end=' : ')
                score = runner.run(*args, **kwargs)
                if verbose > 0:
                    print(score)

                iteration_i += 1
                score_dict[runner.cfg_path] = score
                marks.add(adjacent_code)

                cache.SET('iteration_i', iteration_i)
                cache.SET('current_code', curr)
                cache.SET('marks', marks)
                cache.SET('score_dict', score_dict)
                cache.SAVE()

                if last_score is None or score > last_score:
                    last_score = score
                    found_more = True
                    curr = adjacent_code
                    break



        if not found_more:
            break

    with open(os.path.join(all_cfg_dir, 'seen_scores.json'), 'w') as f:
        json.dump(score_dict, f, indent=4)

    return cache


def grid_search_from_palette(runner: ExperimentRunner,
                             cache_token: str,
                             verbose: int,
                             cfg_base: ConfigurationNode,
                             cfg_palette_dir: str,
                             save_directory: str,
                             *args, **kwargs
                             ):
    cfg_palette_dir = os.path.join(os.getenv('MLM_CONFIG_DIR'), cfg_palette_dir)
    save_directory = os.path.join(os.getenv('MLM_EXPERIMENT_DIR'), save_directory)

    part1, part2 = os.path.split(save_directory)
    sv = os.getcwd()
    os.chdir(part1)
    if not os.path.exists(part2):
        os.mkdir(part2)
    os.chdir(sv)

    # Create a cache in save_directory
    cache = RunnerCache(directory=save_directory, token=cache_token)
    cache.LOAD()
    directories_done = cache.SET_IFN('directories_done', False)
    if not directories_done:
        expand_cfg(cfg_base, cfg_palette_dir, save_directory=save_directory)
    directories_done = True
    cache.SET('directories_done', directories_done)
    cache.SAVE()
    all_cfg_dir = save_directory
    _run_grid_search(runner, cache, verbose, cfg_base, all_cfg_dir, *args, **kwargs)


def grid_search(runner: ExperimentRunner,
                cache_token: str,
                verbose: int,
                cfg_base: ConfigurationNode,
                all_cfg_dir: str,
                *args, **kwargs):
    all_cfg_dir = os.path.join(os.getenv('MLM_EXPERIMENT_DIR'), all_cfg_dir)
    cache = RunnerCache(directory=all_cfg_dir, token=cache_token)
    cache.LOAD()
    _run_grid_search(runner, cache, verbose, cfg_base, all_cfg_dir, *args, **kwargs)
